import asyncio
import sys
from typing import TYPE_CHECKING, Any

from zrb.builtin.llm.chat_session_cmd import (
    ATTACHMENT_CMD,
    HELP_CMD,
    MULTILINE_END_CMD,
    MULTILINE_START_CMD,
    QUIT_CMD,
    RUN_CLI_CMD,
    SAVE_CMD,
    WORKFLOW_CMD,
    YOLO_CMD,
    get_new_attachments,
    get_new_workflows,
    get_new_yolo_mode,
    is_command_match,
    print_commands,
    print_current_attachments,
    print_current_workflows,
    print_current_yolo_mode,
    run_cli_command,
    save_final_result,
)
from zrb.builtin.llm.chat_trigger import llm_chat_trigger
from zrb.config.llm_config import llm_config
from zrb.context.any_context import AnyContext
from zrb.util.cli.markdown import render_markdown

if TYPE_CHECKING:
    from asyncio import StreamReader

    from prompt_toolkit import PromptSession


async def read_user_prompt(ctx: AnyContext) -> str:
    """
    Reads user input from the CLI for an interactive chat session.
    Orchestrates the session by calling helper functions.
    """
    print_commands(ctx)
    is_tty: bool = ctx.is_tty
    reader: PromptSession[Any] | StreamReader = await _setup_input_reader(is_tty)
    multiline_mode = False
    is_first_time = True
    current_workflows: str = ctx.input.workflows
    current_yolo_mode: bool | str = ctx.input.yolo
    current_attachments: str = ctx.input.attach
    user_inputs: list[str] = []
    final_result: str = ""
    should_end = False
    while not should_end:
        await asyncio.sleep(0.01)
        previous_session_name: str | None = (
            ctx.input.previous_session if is_first_time else ""
        )
        start_new: bool = ctx.input.start_new if is_first_time else False
        if is_first_time and ctx.input.message.strip() != "":
            user_input = ctx.input.message
        else:
            # Get user input based on mode
            if not multiline_mode:
                ctx.print("💬 >>", plain=True)
            user_input = await llm_chat_trigger.wait(reader, ctx)
            if not multiline_mode:
                ctx.print("", plain=True)
        # At this point, is_first_time has to be False
        if is_first_time:
            is_first_time = False
        # Handle user input (including slash commands)
        if multiline_mode:
            if is_command_match(user_input, MULTILINE_END_CMD):
                ctx.print("", plain=True)
                multiline_mode = False
            else:
                user_inputs.append(user_input)
                continue
        else:
            if is_command_match(user_input, QUIT_CMD):
                should_end = True
            elif is_command_match(user_input, MULTILINE_START_CMD):
                multiline_mode = True
                ctx.print("", plain=True)
                continue
            elif is_command_match(user_input, WORKFLOW_CMD):
                current_workflows = get_new_workflows(current_workflows, user_input)
                print_current_workflows(ctx, current_workflows)
                continue
            elif is_command_match(user_input, SAVE_CMD):
                save_final_result(ctx, user_input, final_result)
                continue
            elif is_command_match(user_input, ATTACHMENT_CMD):
                current_attachments = get_new_attachments(
                    current_attachments, user_input
                )
                print_current_attachments(ctx, current_attachments)
                continue
            elif is_command_match(user_input, YOLO_CMD):
                current_yolo_mode = get_new_yolo_mode(current_yolo_mode, user_input)
                print_current_yolo_mode(ctx, current_yolo_mode)
                continue
            elif is_command_match(user_input, RUN_CLI_CMD):
                run_cli_command(ctx, user_input)
                continue
            elif is_command_match(user_input, HELP_CMD):
                print_commands(ctx)
                continue
            else:
                user_inputs.append(user_input)
        # Trigger LLM
        user_prompt = "\n".join(user_inputs)
        user_inputs = []
        result = await _trigger_ask_and_wait_for_result(
            ctx=ctx,
            user_prompt=user_prompt,
            attach=current_attachments,
            workflows=current_workflows,
            yolo_mode=current_yolo_mode,
            previous_session_name=previous_session_name,
            start_new=start_new,
        )
        current_attachments = ""
        final_result = final_result if result is None else result
        if ctx.is_web_mode or not is_tty:
            return final_result
    return final_result


async def _setup_input_reader(
    is_interactive: bool,
) -> "PromptSession[Any] | StreamReader":
    """Sets up and returns the appropriate asynchronous input reader."""
    if is_interactive:
        from prompt_toolkit import PromptSession

        return PromptSession()

    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(loop=loop)
    protocol = asyncio.StreamReaderProtocol(reader)
    await loop.connect_read_pipe(lambda: protocol, sys.stdin)
    return reader


async def _trigger_ask_and_wait_for_result(
    ctx: AnyContext,
    user_prompt: str,
    attach: str,
    workflows: str,
    yolo_mode: bool | str,
    previous_session_name: str | None = None,
    start_new: bool = False,
) -> str | None:
    """
    Triggers the LLM ask task and waits for the result via XCom.

    Args:
        ctx: The context object for the task.
        user_prompt: The user's message to send to the LLM.
        previous_session_name: The name of the previous chat session (optional).
        start_new: Whether to start a new conversation (optional).

    Returns:
        The result from the LLM task, or None if the user prompt is empty.
    """
    if user_prompt.strip() == "":
        return None
    await _trigger_ask(
        ctx, user_prompt, attach, workflows, yolo_mode, previous_session_name, start_new
    )
    result = await _wait_ask_result(ctx)
    md_result = render_markdown(result) if result is not None else ""
    ctx.print("\n🤖 >>", plain=True)
    ctx.print(md_result, plain=True)
    ctx.print("", plain=True)
    return result


def get_llm_ask_input_mapping(callback_ctx: AnyContext):
    """
    Generates the input mapping for the LLM ask task from the callback context.

    Args:
        callback_ctx: The context object for the callback.

    Returns:
        A dictionary containing the input mapping for the LLM ask task.
    """
    data = callback_ctx.xcom.ask_trigger.pop()
    system_prompt = callback_ctx.input.system_prompt
    if system_prompt is None or system_prompt.strip() == "":
        system_prompt = llm_config.default_interactive_system_prompt
    return {
        "model": callback_ctx.input.model,
        "base-url": callback_ctx.input.base_url,
        "api-key": callback_ctx.input.api_key,
        "system-prompt": system_prompt,
        "start-new": data.get("start_new"),
        "previous-session": data.get("previous_session_name"),
        "message": data.get("message"),
        "attach": data.get("attach"),
        "workflows": data.get("workflows"),
        "yolo": data.get("yolo"),
    }


async def _trigger_ask(
    ctx: AnyContext,
    user_prompt: str,
    attach: str,
    workflows: str,
    yolo_mode: bool | str,
    previous_session_name: str | None = None,
    start_new: bool = False,
):
    """
    Triggers the LLM ask task by pushing data to the 'ask_trigger' XCom queue.

    Args:
        ctx: The context object for the task.
        user_prompt: The user's message to send to the LLM.
        previous_session_name: The name of the previous chat session (optional).
        start_new: Whether to start a new conversation (optional).
    """
    if previous_session_name is None:
        previous_session_name = await _wait_ask_session_name(ctx)
    ctx.xcom["ask_trigger"].push(
        {
            "previous_session_name": previous_session_name,
            "start_new": start_new,
            "message": user_prompt,
            "attach": attach,
            "workflows": workflows,
            "yolo": yolo_mode,
        }
    )


async def _wait_ask_result(ctx: AnyContext) -> str | None:
    """
    Waits for and retrieves the LLM task result from the 'ask_result' XCom queue.

    Args:
        ctx: The context object for the task.

    Returns:
        The result string from the LLM task.
    """
    while "ask_result" not in ctx.xcom or len(ctx.xcom.ask_result) == 0:
        await asyncio.sleep(0.1)
        if "ask_error" in ctx.xcom and len(ctx.xcom.ask_error) > 0:
            ctx.xcom.ask_error.pop()
            return None
    return ctx.xcom.ask_result.pop()


async def _wait_ask_session_name(ctx: AnyContext) -> str:
    """
    Waits for and retrieves the LLM chat session name from the 'ask_session_name' XCom queue.

    Args:
        ctx: The context object for the task.

    Returns:
        The session name string.
    """
    while "ask_session_name" not in ctx.xcom or len(ctx.xcom.ask_session_name) == 0:
        await asyncio.sleep(0.1)
    return ctx.xcom.ask_session_name.pop()



from my_app_name.common.db_engine_factory import db_engine
from my_app_name.config import APP_REPOSITORY_TYPE
from my_app_name.module.auth.service.user.repository.user_db_repository import (
    UserDBRepository,
)
from my_app_name.module.auth.service.user.repository.user_repository import (
    UserRepository,
)

if APP_REPOSITORY_TYPE == "db":
    user_repository: UserRepository = UserDBRepository(db_engine)
else:
    user_repository: UserRepository = None



from abc import ABC, abstractmethod

from my_app_name.schema.role import (
    Role,
    RoleCreateWithAudit,
    RoleResponse,
    RoleUpdateWithAudit,
)


class RoleRepository(ABC):

    @abstractmethod
    async def get_by_id(self, id: str) -> RoleResponse:
        """Get role by id"""

    @abstractmethod
    async def get_by_ids(self, id_list: list[str]) -> RoleResponse:
        """Get roles by ids"""

    @abstractmethod
    async def validate_permission_names(self, permission_names: list[str]):
        """Validate Permission names"""

    @abstractmethod
    async def add_permissions(self, data: dict[str, list[str]], created_by: str):
        """Adding permissions to roles"""

    @abstractmethod
    async def remove_all_permissions(self, role_ids: list[str] = []):
        """Remove permissions from roles"""

    @abstractmethod
    async def get(
        self,
        page: int = 1,
        page_size: int = 10,
        filter: str | None = None,
        sort: str | None = None,
    ) -> list[Role]:
        """Get roles by filter and sort"""

    @abstractmethod
    async def count(self, filter: str | None = None) -> int:
        """Count roles by filter"""

    @abstractmethod
    async def create(self, data: RoleCreateWithAudit) -> Role:
        """Create a new role"""

    @abstractmethod
    async def create_bulk(self, data: list[RoleCreateWithAudit]) -> list[Role]:
        """Create some roles"""

    @abstractmethod
    async def delete(self, id: str) -> Role:
        """Delete a role"""

    @abstractmethod
    async def delete_bulk(self, id_list: list[str]) -> list[Role]:
        """Delete some roles"""

    @abstractmethod
    async def update(self, id: str, data: RoleUpdateWithAudit) -> Role:
        """Update a role"""

    @abstractmethod
    async def update_bulk(
        self, id_list: list[str], data: RoleUpdateWithAudit
    ) -> list[Role]:
        """Update some roles"""



import importlib.metadata as metadata
import logging
import os
import platform

from zrb.util.string.conversion import to_boolean
from zrb.util.string.format import fstring_format

_DEFAULT_BANNER = """
                bb
   zzzzz rr rr  bb
     zz  rrr  r bbbbbb
    zz   rr     bb   bb
   zzzzz rr     bbbbbb   {VERSION} Janggala
   _ _ . .  . _ .  _ . . .
Your Automation Powerhouse
☕ Donate at: https://stalchmst.com
🐙 Submit issues/PR at: https://github.com/state-alchemists/zrb
🐤 Follow us at: https://twitter.com/zarubastalchmst
"""


class Config:
    def __init__(self):
        self.__internal_default_prompt: dict[str, str] = {}

    @property
    def ENV_PREFIX(self) -> str:
        return os.getenv("_ZRB_ENV_PREFIX", "ZRB")

    def _getenv(self, env_name: str, default: str = "") -> str:
        return os.getenv(f"{self.ENV_PREFIX}_{env_name}", default)

    def _get_internal_default_prompt(self, name: str) -> str:
        if name not in self.__internal_default_prompt:
            file_path = os.path.join(
                os.path.dirname(__file__), "default_prompt", f"{name}.md"
            )
            with open(file_path, "r") as f:
                self.__internal_default_prompt[name] = f.read().strip()
        return self.__internal_default_prompt[name]

    @property
    def LOGGER(self) -> logging.Logger:
        return logging.getLogger()

    @property
    def DEFAULT_SHELL(self) -> str:
        return self._getenv("SHELL", self._get_current_shell())

    def _get_current_shell(self) -> str:
        if platform.system() == "Windows":
            return "PowerShell"
        current_shell = os.getenv("SHELL", "")
        if current_shell.endswith("zsh"):
            return "zsh"
        return "bash"

    @property
    def DEFAULT_EDITOR(self) -> str:
        return self._getenv("EDITOR", "nano")

    @property
    def DEFAULT_DIFF_EDIT_COMMAND_TPL(self) -> str:
        return self._getenv("DIFF_EDIT_COMMAND", self._get_default_diff_edit_command())

    def _get_default_diff_edit_command(self) -> str:
        editor = self.DEFAULT_EDITOR
        if editor in [
            "code",
            "vscode",
            "vscodium",
            "windsurf",
            "cursor",
            "zed",
            "zeditor",
            "agy",
        ]:
            return f"{editor} --wait --diff {{old}} {{new}}"
        if editor == "emacs":
            return 'emacs --eval \'(ediff-files "{old}" "{new}")\''
        if editor in ["nvim", "vim"]:
            return (
                f"{editor} -d {{old}} {{new}} "
                "-i NONE "
                '-c "wincmd h | set readonly | wincmd l" '
                '-c "highlight DiffAdd cterm=bold ctermbg=22 guibg=#005f00 | highlight DiffChange cterm=bold ctermbg=24 guibg=#005f87 | highlight DiffText ctermbg=21 guibg=#0000af | highlight DiffDelete ctermbg=52 guibg=#5f0000" '  # noqa
                '-c "set showtabline=2 | set tabline=[Instructions]\\ :wqa(save\\ &\\ quit)\\ \\|\\ i/esc(toggle\\ edit\\ mode)" '  # noqa
                '-c "wincmd h | setlocal statusline=OLD\\ FILE" '
                '-c "wincmd l | setlocal statusline=%#StatusBold#NEW\\ FILE\\ :wqa(save\\ &\\ quit)\\ \\|\\ i/esc(toggle\\ edit\\ mode)" '  # noqa
                '-c "autocmd BufWritePost * wqa"'
            )
        return 'vimdiff {old} {new} +"setlocal ro" +"wincmd l" +"autocmd BufWritePost <buffer> qa"'  # noqa

    @property
    def INIT_MODULES(self) -> list[str]:
        init_modules_str = self._getenv("INIT_MODULES", "")
        if init_modules_str != "":
            return [
                module.strip()
                for module in init_modules_str.split(":")
                if module.strip() != ""
            ]
        return []

    @property
    def ROOT_GROUP_NAME(self) -> str:
        return self._getenv("ROOT_GROUP_NAME", "zrb")

    @property
    def ROOT_GROUP_DESCRIPTION(self) -> str:
        return self._getenv("ROOT_GROUP_DESCRIPTION", "Your Automation Powerhouse")

    @property
    def INIT_SCRIPTS(self) -> list[str]:
        init_scripts_str = self._getenv("INIT_SCRIPTS", "")
        if init_scripts_str != "":
            return [
                script.strip()
                for script in init_scripts_str.split(":")
                if script.strip() != ""
            ]
        return []

    @property
    def INIT_FILE_NAME(self) -> str:
        return self._getenv("INIT_FILE_NAME", "zrb_init.py")

    @property
    def LOGGING_LEVEL(self) -> int:
        return self._get_log_level(self._getenv("LOGGING_LEVEL", "WARNING"))

    def _get_log_level(self, level: str) -> int:
        level = level.upper()
        log_levels = {
            "CRITICAL": logging.CRITICAL,  # 50
            "FATAL": logging.CRITICAL,  # 50
            "ERROR": logging.ERROR,  # 40
            "WARN": logging.WARNING,  # 30
            "WARNING": logging.WARNING,  # 30
            "INFO": logging.INFO,  # 20
            "DEBUG": logging.DEBUG,  # 10
            "NOTSET": logging.NOTSET,  # 0
        }
        if level in log_levels:
            return log_levels[level]
        return logging.WARNING

    @property
    def LOAD_BUILTIN(self) -> bool:
        return to_boolean(self._getenv("LOAD_BUILTIN", "1"))

    @property
    def WARN_UNRECOMMENDED_COMMAND(self) -> bool:
        return to_boolean(self._getenv("WARN_UNRECOMMENDED_COMMAND", "1"))

    @property
    def SESSION_LOG_DIR(self) -> str:
        return os.getenv(
            "ZRB_SESSION_LOG_DIR", os.path.expanduser(os.path.join("~", ".zrb-session"))
        )

    @property
    def TODO_DIR(self) -> str:
        return self._getenv("TODO_DIR", os.path.expanduser(os.path.join("~", "todo")))

    @property
    def TODO_VISUAL_FILTER(self) -> str:
        return self._getenv("TODO_FILTER", "")

    @property
    def TODO_RETENTION(self) -> str:
        return self._getenv("TODO_RETENTION", "2w")

    @property
    def VERSION(self) -> str:
        custom_version = os.getenv("_ZRB_CUSTOM_VERSION", "")
        if custom_version != "":
            return custom_version
        return metadata.version("zrb")

    @property
    def WEB_CSS_PATH(self) -> list[str]:
        web_css_path_str = self._getenv("WEB_CSS_PATH", "")
        if web_css_path_str != "":
            return [
                path.strip()
                for path in web_css_path_str.split(":")
                if path.strip() != ""
            ]
        return []

    @property
    def WEB_JS_PATH(self) -> list[str]:
        web_js_path_str = self._getenv("WEB_JS_PATH", "")
        if web_js_path_str != "":
            return [
                path.strip()
                for path in web_js_path_str.split(":")
                if path.strip() != ""
            ]
        return []

    @property
    def WEB_FAVICON_PATH(self) -> str:
        return self._getenv("WEB_FAVICON_PATH", "/static/favicon-32x32.png")

    @property
    def WEB_COLOR(self) -> str:
        return self._getenv("WEB_COLOR", "")

    @property
    def WEB_HTTP_PORT(self) -> int:
        return int(self._getenv("WEB_HTTP_PORT", "21213"))

    @property
    def WEB_GUEST_USERNAME(self) -> str:
        return self._getenv("WEB_GUEST_USERNAME", "user")

    @property
    def WEB_SUPER_ADMIN_USERNAME(self) -> str:
        return self._getenv("WEB_SUPER_ADMIN_USERNAME", "admin")

    @property
    def WEB_SUPER_ADMIN_PASSWORD(self) -> str:
        return self._getenv("WEB_SUPER_ADMIN_PASSWORD", "admin")

    @property
    def WEB_ACCESS_TOKEN_COOKIE_NAME(self) -> str:
        return self._getenv("WEB_ACCESS_TOKEN_COOKIE_NAME", "access_token")

    @property
    def WEB_REFRESH_TOKEN_COOKIE_NAME(self) -> str:
        return self._getenv("WEB_REFRESH_TOKEN_COOKIE_NAME", "refresh_token")

    @property
    def WEB_SECRET_KEY(self) -> str:
        return self._getenv("WEB_SECRET", "zrb")

    @property
    def WEB_ENABLE_AUTH(self) -> bool:
        return to_boolean(self._getenv("WEB_ENABLE_AUTH", "0"))

    @property
    def WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES(self) -> int:
        return int(self._getenv("WEB_ACCESS_TOKEN_EXPIRE_MINUTES", "30"))

    @property
    def WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES(self) -> int:
        return int(self._getenv("WEB_REFRESH_TOKEN_EXPIRE_MINUTES", "60"))

    @property
    def WEB_TITLE(self) -> str:
        return self._getenv("WEB_TITLE", "Zrb")

    @property
    def WEB_JARGON(self) -> str:
        return self._getenv("WEB_JARGON", "Your Automation PowerHouse")

    @property
    def WEB_HOMEPAGE_INTRO(self) -> str:
        return self._getenv("WEB_HOMEPAGE_INTRO", "Welcome to Zrb Web Interface")

    @property
    def LLM_MODEL(self) -> str | None:
        value = self._getenv("LLM_MODEL")
        return None if value == "" else value

    @property
    def LLM_BASE_URL(self) -> str | None:
        value = self._getenv("LLM_BASE_URL")
        return None if value == "" else value

    @property
    def LLM_API_KEY(self) -> str | None:
        value = self._getenv("LLM_API_KEY")
        return None if value == "" else value

    @property
    def LLM_MODEL_SMALL(self) -> str | None:
        value = self._getenv("LLM_MODEL_SMALL")
        return None if value == "" else value

    @property
    def LLM_BASE_URL_SMALL(self) -> str | None:
        value = self._getenv("LLM_BASE_URL_SMALL")
        return None if value == "" else value

    @property
    def LLM_API_KEY_SMALL(self) -> str | None:
        value = self._getenv("LLM_API_KEY_SMALL")
        return None if value == "" else value

    @property
    def LLM_SYSTEM_PROMPT(self) -> str | None:
        value = self._getenv("LLM_SYSTEM_PROMPT")
        return None if value == "" else value

    @property
    def LLM_INTERACTIVE_SYSTEM_PROMPT(self) -> str | None:
        value = self._getenv("LLM_INTERACTIVE_SYSTEM_PROMPT")
        return None if value == "" else value

    @property
    def LLM_PERSONA(self) -> str | None:
        value = self._getenv("LLM_PERSONA")
        return None if value == "" else value

    @property
    def LLM_WORKFLOWS(self) -> list[str]:
        """Get a list of LLM workflows from environment variables."""
        workflows = []
        for workflow in self._getenv("LLM_WORKFLOWS", "").split(","):
            workflow = workflow.strip()
            if workflow != "":
                workflows.append(workflow)
        return workflows

    @property
    def LLM_BUILTIN_WORKFLOW_PATHS(self) -> list[str]:
        """Get a list of additional builtin workflow paths from environment variables."""
        builtin_workflow_paths_str = self._getenv("LLM_BUILTIN_WORKFLOW_PATHS", "")
        if builtin_workflow_paths_str != "":
            return [
                path.strip()
                for path in builtin_workflow_paths_str.split(":")
                if path.strip() != ""
            ]
        return []

    @property
    def LLM_SPECIAL_INSTRUCTION_PROMPT(self) -> str | None:
        value = self._getenv("LLM_SPECIAL_INSTRUCTION_PROMPT")
        return None if value == "" else value

    @property
    def LLM_SUMMARIZATION_PROMPT(self) -> str | None:
        value = self._getenv("LLM_SUMMARIZATION_PROMPT")
        return None if value == "" else value

    @property
    def LLM_MAX_REQUESTS_PER_MINUTE(self) -> int:
        """
        Maximum number of LLM requests allowed per minute.
        Default is conservative to accommodate free-tier LLM providers.
        """
        return int(self._getenv("LLM_MAX_REQUESTS_PER_MINUTE", "60"))

    @property
    def LLM_MAX_TOKENS_PER_MINUTE(self) -> int:
        """
        Maximum number of LLM tokens allowed per minute.
        Default is conservative to accommodate free-tier LLM providers.
        """
        return int(self._getenv("LLM_MAX_TOKENS_PER_MINUTE", "100000"))

    @property
    def LLM_MAX_TOKENS_PER_REQUEST(self) -> int:
        """Maximum number of tokens allowed per individual LLM request."""
        return int(self._getenv("LLM_MAX_TOKENS_PER_REQUEST", "100000"))

    @property
    def LLM_MAX_TOKENS_PER_TOOL_CALL_RESULT(self) -> int:
        """Maximum number of tokens allowed per tool call result."""
        return int(self._getenv("LLM_MAX_TOKENS_PER_TOOL_CALL_RESULT", "75000"))

    @property
    def LLM_THROTTLE_SLEEP(self) -> float:
        """Number of seconds to sleep when throttling is required."""
        return float(self._getenv("LLM_THROTTLE_SLEEP", "5.0"))

    @property
    def LLM_YOLO_MODE(self) -> bool | list[str]:
        str_val = self._getenv("LLM_YOLO_MODE", "false")
        try:
            return to_boolean(str_val)
        except Exception:
            return [val.strip() for val in str_val.split(",") if val.strip() != ""]

    @property
    def LLM_SUMMARIZE_HISTORY(self) -> bool:
        return to_boolean(self._getenv("LLM_SUMMARIZE_HISTORY", "true"))

    @property
    def LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD(self) -> int:
        threshold = int(
            self._getenv("LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD", "20000")
        )
        return self._limit_token_threshold(threshold)

    @property
    def LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_THRESHOLD(self) -> int:
        threshold = int(
            self._getenv("LLM_REPO_ANALYSIS_EXTRACTION_TOKEN_LIMIT", "50000")
        )
        return self._limit_token_threshold(threshold)

    @property
    def LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_THRESHOLD(self) -> int:
        threshold = int(
            self._getenv("LLM_REPO_ANALYSIS_SUMMARIZATION_TOKEN_LIMIT", "20000")
        )
        min(
            threshold,
            self.LLM_MAX_TOKENS_PER_MINUTE // 2,
            self.LLM_MAX_TOKENS_PER_REQUEST // 2,
        )
        return self._limit_token_threshold(threshold)

    @property
    def LLM_FILE_ANALYSIS_TOKEN_LIMIT(self) -> int:
        threshold = int(self._getenv("LLM_FILE_ANALYSIS_TOKEN_LIMIT", "50000"))
        return self._limit_token_threshold(threshold)

    def _limit_token_threshold(self, threshold: int) -> int:
        return min(
            threshold,
            self.LLM_MAX_TOKENS_PER_MINUTE // 2,
            self.LLM_MAX_TOKENS_PER_REQUEST // 2,
        )

    @property
    def LLM_FILE_EXTRACTOR_SYSTEM_PROMPT(self) -> str:
        return self._getenv(
            "LLM_FILE_EXTRACTOR_SYSTEM_PROMPT",
            self._get_internal_default_prompt("file_extractor_system_prompt"),
        )

    @property
    def LLM_REPO_EXTRACTOR_SYSTEM_PROMPT(self) -> str:
        return self._getenv(
            "LLM_REPO_EXTRACTOR_SYSTEM_PROMPT",
            self._get_internal_default_prompt("repo_extractor_system_prompt"),
        )

    @property
    def LLM_REPO_SUMMARIZER_SYSTEM_PROMPT(self) -> str:
        return self._getenv(
            "LLM_REPO_SUMMARIZER_SYSTEM_PROMPT",
            self._get_internal_default_prompt("repo_summarizer_system_prompt"),
        )

    @property
    def LLM_HISTORY_DIR(self) -> str:
        return self._getenv(
            "LLM_HISTORY_DIR",
            os.path.expanduser(os.path.join("~", ".zrb-llm-history")),
        )

    @property
    def LLM_ALLOW_ACCESS_LOCAL_FILE(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_ACCESS_LOCAL_FILE", "1"))

    @property
    def LLM_ALLOW_ANALYZE_FILE(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_ANALYZE_LOCAL_FILE", "1"))

    @property
    def LLM_ALLOW_ANALYZE_REPO(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_ANALYZE_REPO", "1"))

    @property
    def LLM_ALLOW_ACCESS_SHELL(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_ACCESS_SHELL", "1"))

    @property
    def LLM_ALLOW_OPEN_WEB_PAGE(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_OPEN_WEB_PAGE", "1"))

    @property
    def LLM_ALLOW_SEARCH_INTERNET(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_SEARCH_INTERNET", "1"))

    @property
    def LLM_ALLOW_GET_CURRENT_LOCATION(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_GET_CURRENT_LOCATION", "1"))

    @property
    def LLM_ALLOW_GET_CURRENT_WEATHER(self) -> bool:
        return to_boolean(self._getenv("LLM_ALLOW_GET_CURRENT_WEATHER", "1"))

    @property
    def RAG_EMBEDDING_API_KEY(self) -> str | None:
        value = self._getenv("RAG_EMBEDDING_API_KEY")
        return None if value == "" else value

    @property
    def RAG_EMBEDDING_BASE_URL(self) -> str | None:
        value = self._getenv("RAG_EMBEDDING_BASE_URL")
        return None if value == "" else value

    @property
    def RAG_EMBEDDING_MODEL(self) -> str:
        return self._getenv("RAG_EMBEDDING_MODEL", "text-embedding-ada-002")

    @property
    def RAG_CHUNK_SIZE(self) -> int:
        return int(self._getenv("RAG_CHUNK_SIZE", "1024"))

    @property
    def RAG_OVERLAP(self) -> int:
        return int(self._getenv("RAG_OVERLAP", "128"))

    @property
    def RAG_MAX_RESULT_COUNT(self) -> int:
        return int(self._getenv("RAG_MAX_RESULT_COUNT", "5"))

    @property
    def SEARCH_INTERNET_METHOD(self) -> str:
        """Either serpapi or searxng"""
        return self._getenv("SEARCH_INTERNET_METHOD", "serpapi")

    @property
    def BRAVE_API_KEY(self) -> str:
        return os.getenv("BRAVE_API_KEY", "")

    @property
    def BRAVE_API_SAFE(self) -> str:
        return self._getenv("BRAVE_API_SAFE", "off")

    @property
    def BRAVE_API_LANG(self) -> str:
        return self._getenv("BRAVE_API_LANG", "en")

    @property
    def SERPAPI_KEY(self) -> str:
        return os.getenv("SERPAPI_KEY", "")

    @property
    def SERPAPI_SAFE(self) -> str:
        return self._getenv("SERPAPI_SAFE", "off")

    @property
    def SERPAPI_LANG(self) -> str:
        return self._getenv("SERPAPI_LANG", "en")

    @property
    def SEARXNG_PORT(self) -> int:
        return int(self._getenv("SEARXNG_PORT", "8080"))

    @property
    def SEARXNG_BASE_URL(self) -> str:
        return self._getenv("SEARXNG_BASE_URL", f"http://localhost:{self.SEARXNG_PORT}")

    @property
    def SEARXNG_SAFE(self) -> int:
        return int(self._getenv("SEARXNG_SAFE", "0"))

    @property
    def SEARXNG_LANG(self) -> str:
        return self._getenv("SEARXNG_LANG", "en")

    @property
    def BANNER(self) -> str:
        return fstring_format(
            self._getenv("BANNER", _DEFAULT_BANNER),
            {"VERSION": self.VERSION},
        )

    @property
    def LLM_CONTEXT_FILE(self) -> str:
        return self._getenv("LLM_CONTEXT_FILE", "ZRB.md")

    @property
    def USE_TIKTOKEN(self) -> bool:
        return to_boolean(self._getenv("USE_TIKTOKEN", "true"))

    @property
    def TIKTOKEN_ENCODING_NAME(self) -> str:
        return self._getenv("TIKTOKEN_ENCODING_NAME", "cl100k_base")


CFG = Config()



import os

from zrb.builtin.group import project_group
from zrb.input.str_input import StrInput
from zrb.task.scaffolder import Scaffolder
from zrb.task.task import Task

_DIR = os.path.dirname(__file__)

scaffold_project = Scaffolder(
    name="scaffold-project",
    description="🌟 Create project",
    input=[
        StrInput(
            name="project-dir",
            description="Project directory",
            prompt="Project directory",
            default=lambda _: os.getcwd(),
        ),
        StrInput(
            name="project",
            description="Project name",
            prompt="Project name",
            default=lambda ctx: os.path.basename(ctx.input.project_dir),
        ),
    ],
    source_path=os.path.join(_DIR, "project-template"),
    render_source_path=False,
    destination_path="{ctx.input['project-dir']}",
    transform_content={"Project Name": "{ctx.input.project.title()}"},
    retries=0,
)

create_project = project_group.add_task(
    Task(
        name="create-project",
        description="🌟 Create project",
    ),
    alias="create",
)
scaffold_project >> create_project



import json
import os

from zrb.config.config import CFG
from zrb.group.any_group import AnyGroup
from zrb.runner.web_schema.user import User
from zrb.runner.web_util.html import get_html_auth_link
from zrb.session.any_session import AnySession
from zrb.task.any_task import AnyTask
from zrb.util.file import read_file
from zrb.util.string.format import fstring_format


def show_task_page(
    user: User,
    root_group: AnyGroup,
    task: AnyTask,
    session: AnySession,
    url: str,
    args: list[str],
):
    from fastapi.responses import HTMLResponse

    _DIR = os.path.dirname(__file__)
    _GLOBAL_TEMPLATE = read_file(
        os.path.join(
            os.path.dirname(os.path.dirname(_DIR)), "static", "global_template.html"
        )
    )
    _VIEW_TEMPLATE = read_file(os.path.join(_DIR, "view.html"))
    _TASK_INPUT_TEMPLATE = read_file(os.path.join(_DIR, "partial", "input.html"))
    web_title = CFG.WEB_TITLE if CFG.WEB_TITLE.strip() != "" else root_group.name
    web_jargon = (
        CFG.WEB_JARGON if CFG.WEB_JARGON.strip() != "" else root_group.description
    )
    auth_link = get_html_auth_link(user)
    session.register_task(task)
    ctx = task.get_ctx(session)
    url_parts = url.split("/")
    # Assemble parent url
    parent_url_parts = url_parts[:-2] + [""]
    parent_url = "/".join(parent_url_parts)
    # Assemble session api url
    session_url_parts = list(url_parts)
    session_url_parts[1] = "api/v1/task-sessions"
    session_api_url = "/".join(session_url_parts)
    # Assemble input api url
    input_url_parts = list(url_parts)
    input_url_parts[1] = "api/v1/task-inputs"
    input_api_url = "/".join(input_url_parts)
    # Assemble ui url
    ui_url_parts = list(url_parts)
    ui_url_parts[1] = "ui"
    ui_url = "/".join(ui_url_parts)
    # Assemble task inputs
    input_html_list = []
    for task_input in task.inputs:
        task_input.update_shared_context(ctx)
        input_html_list.append(
            fstring_format(_TASK_INPUT_TEMPLATE, {"task_input": task_input, "ctx": ctx})
        )
    session_name = args[0] if len(args) > 0 else ""
    return HTMLResponse(
        fstring_format(
            _GLOBAL_TEMPLATE,
            {
                "CFG": CFG,
                "root_group": root_group,
                "content": fstring_format(
                    _VIEW_TEMPLATE,
                    {
                        "web_title": web_title,
                        "web_jargon": web_jargon,
                        "name": task.name,
                        "description": task.description,
                        "auth_link": auth_link,
                        "url": url,
                        "parent_url": parent_url,
                        "task_inputs": "\n".join(input_html_list),
                        "ui_url": ui_url,
                        "json_cfg": json.dumps(
                            {
                                "CURRENT_URL": url,
                                "SESSION_API_URL": session_api_url,
                                "INPUT_API_URL": input_api_url,
                                "UI_URL": ui_url,
                                "SESSION_NAME": session_name,
                                "PAGE": 0,
                            }
                        ),
                    },
                ),
            },
        )
    )



import os
import re
from typing import Literal


def read_file(file_path: str, replace_map: dict[str, str] = {}) -> str:
    """Reads a file and optionally replaces content based on a map.

    Args:
        file_path: The path to the file.
        replace_map: A dictionary of strings to replace.

    Returns:
        The content of the file with replacements applied.
    """
    abs_file_path = os.path.abspath(os.path.expanduser(file_path))
    is_pdf = abs_file_path.lower().endswith(".pdf")
    try:
        content = (
            _read_pdf_file_content(abs_file_path)
            if is_pdf
            else _read_text_file_content(abs_file_path)
        )
        for key, val in replace_map.items():
            content = content.replace(key, val)
        return content
    except Exception:
        import base64
        from pathlib import Path

        data = Path(abs_file_path).read_bytes()
        return base64.b64encode(data).decode("ascii")


def _read_text_file_content(file_path: str) -> str:
    with open(file_path, "r", encoding="utf-8") as f:
        content = f.read()
    return content


def _read_pdf_file_content(file_path: str) -> str:
    import pdfplumber
    from pdfplumber.pdf import PDF

    with pdfplumber.open(file_path) as pdf:
        pdf: PDF
        return "\n".join(
            page.extract_text() for page in pdf.pages if page.extract_text()
        )


def read_file_with_line_numbers(
    file_path: str, replace_map: dict[str, str] = {}
) -> str:
    """Reads a file and returns content with line numbers.

    Args:
        file_path: The path to the file.
        replace_map: A dictionary of strings to replace.

    Returns:
        The content of the file with line numbers and replacements applied.
    """
    content = read_file(file_path, replace_map)
    if not content:
        return ""
    lines = content.splitlines()
    numbered_lines = [f"{i + 1} | {line}" for i, line in enumerate(lines)]
    return "\n".join(numbered_lines)


def read_dir(dir_path: str) -> list[str]:
    """Reads a directory and returns a list of file names.

    Args:
        dir_path: The path to the directory.

    Returns:
        A list of file names in the directory.
    """
    return [f for f in os.listdir(os.path.abspath(os.path.expanduser(dir_path)))]


def write_file(
    file_path: str,
    content: str | list[str],
    mode: Literal["w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx"] = "w",
):
    """Writes content to a file.

    Args:
        file_path: The path to the file.
        content: The content to write, either a string or a list of strings.
        mode: Writing mode (by default "w")
    """
    if isinstance(content, list):
        content = "\n".join([line for line in content if line is not None])
    abs_file_path = os.path.abspath(os.path.expanduser(file_path))
    dir_path = os.path.dirname(abs_file_path)
    os.makedirs(dir_path, exist_ok=True)
    should_add_eol = content.endswith("\n")
    # Remove trailing newlines, but keep one if the file originally ended up with newline
    content = re.sub(r"\n{3,}$", "\n\n", content)
    content = content.rstrip("\n")
    if should_add_eol:
        content += "\n"
    with open(abs_file_path, mode) as f:
        f.write(content)



from typing import Callable

from zrb.task.llm.conversation_history_model import ConversationHistory


def create_history_summarization_tool(
    conversation_history: ConversationHistory,
) -> Callable[[str, str], str]:
    def update_conversation_memory(
        past_conversation_summary: str,
        past_conversation_transcript: str,
    ) -> str:
        """
        Update the conversation memory including summary and transcript.
        - past_conversation_summary: A concise narrative that integrates the
          previous summary with the recent conversation.
        - past_conversation_transcript: MUST be ONLY the last 4 (four) turns
          of the conversation.
        """
        conversation_history.past_conversation_summary = past_conversation_summary
        conversation_history.past_conversation_transcript = past_conversation_transcript
        return "Conversation memory updated"

    return update_conversation_memory



import os

from zrb.config.llm_context.config import llm_context_config


def read_long_term_note() -> str:
    """
    Retrieves the GLOBAL long-term memory shared across ALL sessions and projects.

    CRITICAL: Consult this first for user preferences, facts, and cross-project context.

    Returns:
        str: The current global note content.
    """
    contexts = llm_context_config.get_notes()
    return contexts.get("/", "")


def write_long_term_note(content: str) -> str:
    """
    Persists CRITICAL facts to the GLOBAL long-term memory.

    USE EAGERLY to save:
    - User preferences (e.g., "I prefer Python", "No unit tests").
    - User information (e.g., user name, user email address).
    - Important facts (e.g., "My API key is in .env").
    - Cross-project goals.
    - Anything that will be useful for future interaction across projects.

    WARNING: This OVERWRITES the entire global note. Always read first.

    Args:
        content (str): The text to strictly memorize.

    Returns:
        str: Confirmation message.
    """
    llm_context_config.write_note(content, "/")
    return "Global long-term note saved."


def read_contextual_note(path: str | None = None) -> str:
    """
    Retrieves LOCAL memory specific to a file or directory path.

    Use to recall project-specific architecture, code summaries, or past decisions
    relevant to the current working location.

    Args:
        path (str | None): Target file/dir. Defaults to current working directory (CWD).

    Returns:
        str: The local note content for the path.
    """
    if path is None:
        path = os.getcwd()
    abs_path = os.path.abspath(path)
    contexts = llm_context_config.get_notes(cwd=abs_path)
    return contexts.get(abs_path, "")


def write_contextual_note(content: str, path: str | None = None) -> str:
    """
    Persists LOCAL facts specific to a file or directory.

    USE EAGERLY to save:
    - Architectural patterns for this project/directory.
    - Summaries of large files or directories.
    - Specific guidelines for this project.
    - Anything related to this directory that will be useful for future interaction.

    WARNING: This OVERWRITES the note for the specific path. Always read first.

    Args:
        content (str): The text to memorize for this location.
        path (str | None): Target file/dir. Defaults to CWD.

    Returns:
        str: Confirmation message.
    """
    if path is None:
        path = os.getcwd()
    llm_context_config.write_note(content, path)
    return f"Contextual note saved for: {path}"



import os

from my_app_name._zrb.config import (
    ACTIVATE_VENV_SCRIPT,
    APP_DIR,
    MICROSERVICES_ENV_VARS,
    MONOLITH_ENV_VARS,
)
from my_app_name._zrb.input import run_env_input
from my_app_name._zrb.util import (
    cd_module_script,
    run_my_app_name,
    set_create_migration_db_url_env,
    set_env,
    set_module_env,
)

from zrb import Cmd, CmdTask, EnvFile, EnvMap, StrInput, Task
from zrb.util.string.conversion import to_kebab_case, to_snake_case


def create_migration(module: str) -> Task:
    name = to_kebab_case(module)
    return CmdTask(
        name=f"create-my-app-name-{name}-migration",
        description=f"🧩 Create My App Name {name.capitalize()} DB migration",
        input=StrInput(
            name="message",
            description="Migration message",
            prompt="Migration message",
            allow_empty=False,
        ),
        env=EnvFile(path=os.path.join(APP_DIR, "template.env")),
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            set_create_migration_db_url_env(module),
            set_module_env(module),
            cd_module_script(module),
            "alembic upgrade head",
            Cmd(
                "alembic revision --autogenerate -m {double_quote(ctx.input.message)}",
                auto_render=True,
            ),
        ],
        render_cmd=False,
        retries=2,
    )


def migrate_module(
    module: str, as_microservices: bool, additional_env_vars: dict[str, str] = {}
) -> Task:
    name = to_kebab_case(module)
    env_vars = (
        dict(MICROSERVICES_ENV_VARS) if as_microservices else dict(MONOLITH_ENV_VARS)
    )
    env_vars.update(additional_env_vars)
    if as_microservices:
        env_vars["MY_APP_NAME_MODULES"] = to_snake_case(module)
    return CmdTask(
        name=(
            f"migrate-my-app-name-{name}"
            if as_microservices
            else f"migrate-{name}-on-monolith"
        ),
        description=f"🧩 Run My App Name {name.capitalize()} DB migration",
        env=[
            EnvFile(path=os.path.join(APP_DIR, "template.env")),
            EnvMap(vars=env_vars),
        ],
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            cd_module_script(module),
            "alembic upgrade head",
        ],
        render_cmd=False,
        retries=2,
    )


def run_microservice(module: str, port: int) -> Task:
    name = to_kebab_case(module)
    return CmdTask(
        name=f"run-my-app-name-{name}",
        description=f"🧩 Run My App Name {name.capitalize()}",
        input=run_env_input,
        env=[
            EnvFile(path=os.path.join(APP_DIR, "template.env")),
            EnvMap(
                vars={
                    **MICROSERVICES_ENV_VARS,
                }
            ),
        ],
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            set_env("MY_APP_NAME_MODULES", module),
            set_env("MY_APP_NAME_PORT", f"{port}"),
            run_my_app_name,
        ],
        render_cmd=False,
        retries=2,
    )



from fastapi.testclient import TestClient
from my_app_name.main import app
from my_app_name.test._util.access_token import get_admin_access_token


def test_update_permission():
    client = TestClient(app, base_url="http://localhost")
    access_token = get_admin_access_token()
    new_permission_data = {
        "name": "to-be-updated-permission",
        "description": "to-be-updated-permission-description",
    }
    insert_response = client.post(
        "/api/v1/permissions",
        json=new_permission_data,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert insert_response.status_code == 200
    id = insert_response.json().get("id")
    # updating
    updated_permission_data = {
        "name": "updated-permission",
        "description": "updated-permission-description",
    }
    response = client.put(
        f"/api/v1/permissions/{id}",
        json=updated_permission_data,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert response.status_code == 200
    response_data = response.json()
    assert response_data.get("id") == id
    assert response_data.get("name") == "updated-permission"
    assert response_data.get("description") == "updated-permission-description"


def test_update_permission_bulk():
    client = TestClient(app, base_url="http://localhost")
    access_token = get_admin_access_token()
    new_first_permission_data = {
        "name": "to-be-updated-permission-bulk-1",
        "description": "to-be-updated-permission-bulk-description-1",
    }
    insert_response = client.post(
        "/api/v1/permissions/bulk",
        json=[new_first_permission_data],
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert insert_response.status_code == 200
    ids = [row["id"] for row in insert_response.json()]
    # updating (we only test with one data)
    updated_permission_data = {"description": "updated-permission-description"}
    response = client.put(
        f"/api/v1/permissions/bulk",
        json={
            "permission_ids": ids,
            "data": updated_permission_data,
        },
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert response.status_code == 200
    response_data = response.json()
    assert len(response_data) == 1
    response_data[0].get("id") == ids[0]
    response_data[0].get("name") == new_first_permission_data["name"]
    response_data[0].get("description") == new_first_permission_data["description"]



from zrb.config.config import CFG
from zrb.group.group import Group
from zrb.runner.cli import cli


def _maybe_add_group(group: Group):
    if CFG.LOAD_BUILTIN:
        cli.add_group(group)
    return group


base64_group = _maybe_add_group(
    Group(name="base64", description="📄 Base64 operations")
)
uuid_group = _maybe_add_group(Group(name="uuid", description="🆔 UUID operations"))
uuid_v1_group = uuid_group.add_group(Group(name="v1", description="UUID V1 operations"))
uuid_v3_group = uuid_group.add_group(Group(name="v3", description="UUID V3 operations"))
uuid_v4_group = uuid_group.add_group(Group(name="v4", description="UUID V4 operations"))
uuid_v5_group = uuid_group.add_group(Group(name="v5", description="UUID V5 operations"))
ulid_group = _maybe_add_group(Group(name="ulid", description="🔢 ULID operations"))
jwt_group = _maybe_add_group(Group(name="jwt", description="🔒 JWT encode/decode"))
http_group = _maybe_add_group(
    Group(name="http", description="🌐 HTTP request operations")
)

random_group = _maybe_add_group(Group(name="random", description="🔀 Random operation"))
git_group = _maybe_add_group(Group(name="git", description="🌱 Git related commands"))
git_branch_group = git_group.add_group(
    Group(name="branch", description="🌿 Git branch related commands")
)
git_subtree_group = git_group.add_group(
    Group(name="subtree", description="🌳 Git subtree related commands")
)
llm_group = _maybe_add_group(Group(name="llm", description="🤖 LLM operations"))
md5_group = _maybe_add_group(Group(name="md5", description="🔢 Md5 operations"))
python_group = _maybe_add_group(
    Group(name="python", description="🐍 Python related commands")
)
todo_group = _maybe_add_group(Group(name="todo", description="✅ Todo management"))

shell_group = _maybe_add_group(
    Group(name="shell", description="💬 Shell related commands")
)
shell_autocomplete_group = shell_group.add_group(
    Group(name="autocomplete", description="⌨️ Shell autocomplete related commands")
)

project_group = _maybe_add_group(
    Group(name="project", description="📁 Project related commands")
)
add_to_project_group = project_group.add_group(
    Group(name="add", description="➕ Add things to project")
)
add_fastapp_to_project_group = add_to_project_group.add_group(
    Group(name="fastapp", description="🚀 Add Fastapp resources")
)

setup_group = _maybe_add_group(Group(name="setup", description="🔧 Setup"))
setup_latex_group = setup_group.add_group(
    Group(name="latex", description="✍️ Setup LaTeX")
)

searxng_group = _maybe_add_group(
    Group(name="searxng", description="🔎 Searxng related command")
)



from zrb.group.any_group import AnyGroup
from zrb.util.group import get_non_empty_subgroups, get_subtasks


class SubCommand:
    def __init__(self, paths: list[str] = [], nexts: list[str] = []):
        """
        Initialize a SubCommand object.

        Args:
            paths (list[str]): The list of path components leading to this subcommand.
            nexts (list[str]): The list of possible next subcommand or task names.
        """
        self.paths = paths
        self.nexts = nexts

    def __repr__(self):
        """
        Return a string representation of the SubCommand object.
        """
        return f"<{self.__class__.__name__} paths={self.paths} nexts={self.nexts}>"


def get_group_subcommands(
    group: AnyGroup, previous_path: list[str] = [], subcommands: list[SubCommand] = []
) -> list[SubCommand]:
    """
    Recursively get all possible subcommands within a group hierarchy.

    Args:
        group (AnyGroup): The current group to process.
        previous_path (list[str]): The path leading to the current group.
        subcommands (list[SubCommand]): The list to accumulate subcommands.

    Returns:
        list[SubCommand]: A list of all discovered subcommands.
    """
    nexts = []
    for task_alias in get_subtasks(group):
        nexts.append(task_alias)
    for subgroup_alias, subgroup in get_non_empty_subgroups(group).items():
        nexts.append(subgroup_alias)
        # Recursively add subgroup
        get_group_subcommands(
            group=subgroup,
            previous_path=previous_path + [group.name],
            subcommands=subcommands,
        )
    if len(nexts) > 0:
        subcommands.append(SubCommand(paths=previous_path + [group.name], nexts=nexts))
    return subcommands
